/*
 Quasar E1 interface board Linux driver.

 Copyright (C) 2006-2007 Parabel Ltd.
 Author: Mikhail Khelik <mixel.net@gmail.com>

 Copyright (C) 2008-2011 Parabel Ltd.
 Author: Pavel Selivanov <pavel[AT]parabel.ru>

 * Parabel Ltd. reserve the right to distribute project or it's parts
 under the alternative conditions.

 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software Foundation;
 either version 2 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along with this program;
 if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/ioctl.h>
#include <linux/poll.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/smp_lock.h> /* lock_kernel */
#include <linux/version.h>

#include "kal.h"
#include "version.h"
#include "pmc_e1_conf.h"
#include "pmc_e1.h"
#include "e12zap.h"
#include "pmc_e1_if.h" //part of ecfg, include after <linux/ioctl.h>

/* Defines */
#define DRV_NAME "quasar"
#define WIN_SIZE 0x10000 /* PCI window size */

/* Global variables */
static dev_t dev_file_glob;	/* descriptop for devfs and sysfs */
static unsigned char device_number = 0;
static unsigned int cacheline = 0x20;
static unsigned int cacheline_bridge = 0x20;
static unsigned int latency = 0;
static int irqflags = IRQF_SHARED | IRQF_DISABLED;
LIST_HEAD(dev_list_head);

/* Global initialized structs */
static struct pci_device_id ptmc_e1_tbl[] = {
	{ 0x1172, 0x2222, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ 0, }
};
MODULE_DEVICE_TABLE(pci, ptmc_e1_tbl);

/* Function prototypes */
/* */

#ifdef PMC_E1_UDEV
static struct class *pmc_e1_class = NULL;
#else
# define ptmc_sysfs_create(a)
# define ptmc_sysfs_delete(a)
#endif //PMC_E1_UDEV

/*========== Common functions ==========*/
static void init_ptmc_regs(const struct pmc_e1_controller * tmp_dev)
{
	unsigned int output_2048=0;
	unsigned int use_2048_for_sync = 0;
	if (PMC_OLD_FIRMWARE != (tmp_dev->bar0[0x0D] & 0xffff)) //Set to output 2048 kb instead of 8 kb
	{
		output_2048 = PMC_CT_2048_OUT;
	     //   use_2048_for_sync = PMC_USE_FOR_SYNC_2048;
	}
	tmp_dev->bar0[0x02] = 0xc1000000 | MODE1_MA_SL_SEL_HP | PMC_CT_TI_EN | PMC_DMARST | PMC_CONS_DEL | output_2048 | use_2048_for_sync; //Configure board
	udelay(1000);
	tmp_dev->bar0[0x02] = 0x41000000 | PMC_WORKING_MODE | MODE1_FLOCK | MODE1_MA_SL_SEL_HP | PMC_CT_TI_EN | PMC_DMARST | PMC_CONS_DEL | output_2048 | use_2048_for_sync; //Wake up framers, hm magic reset
	udelay(1000);
	tmp_dev->bar0[0x01] = 0x80000000; // Enable interrupts, in general;
	tmp_dev->bar0[0x06] = 0x000000ff; // BS/LS: allow to sync to all E1 channels;
	tmp_dev->bar0[0x04] = 0x000000ff; // TX_enable: enable transmit;
}


static void init_switch_map(const struct pmc_e1_controller * tmp_dev)
{
    int i, j;
    //#define CTB_TO_E1_TABLE(ct_ch, ct_ts)	( ct_ts | ( ct_ch << 8) )
    //#define	E1_TO_CTB_TABLE(e1_ch, e1_ts)	( ( e1_ts | 0x00008000) | (e1_ch  << 8) )

    //DAHDI
    //mapping of timeslots for dahdi
    //PC=>E1
    //DMA_LINE_2 => CT_BUS_4 => E1_LINE0
    //DMA_LINE_3 => CT_BUS_5 => E1_LINE1
    for(i=0;i<128;i++)
    {//LINE2 -> CT_BUS4
        tmp_dev->bar0[0x2400+i] =  E1_TO_CTB_TABLE(2,i);
	tmp_dev->bar0[0x2500+i] =  E1_TO_CTB_TABLE(3,i);
    }
    for(i=0;i < 32;i++)   
    {
	//CT_BUS => E1; according to suggestion.
	//4 CT chanal
	//CT_BUS4 -> LINE0
	tmp_dev->bar0[0x1000+i] = CTB_TO_E1_TABLE(4,i);
	tmp_dev->bar0[0x1100+i] = CTB_TO_E1_TABLE(4,(i + 32));
	tmp_dev->bar0[0x1200+i] = CTB_TO_E1_TABLE(4,(i + 64));
	tmp_dev->bar0[0x1300+i] = CTB_TO_E1_TABLE(4,(i + 96));
	//CT_BUS5 -> LINE1
	tmp_dev->bar0[0x1400+i] = CTB_TO_E1_TABLE(5,i);
	tmp_dev->bar0[0x1500+i] = CTB_TO_E1_TABLE(5,(i + 32));
	tmp_dev->bar0[0x1600+i] = CTB_TO_E1_TABLE(5,(i + 64));
	tmp_dev->bar0[0x1700+i] = CTB_TO_E1_TABLE(5,(i + 96));
    }
    //E1=>PC
    for (i=0;i < 32; i++)
    {
	//E1 => CT_BUS
	for(j=0;j < 4; j++)
	{
	    //E1_LINE0 => CT_BUS_6 => WRITE_BUF_0
	    tmp_dev->bar0[0x2600+i+32*j] =  E1_TO_CTB_TABLE(0,(i*4+j));
	    //E1_LINE1 => CT_BUS_7 => WRITE_BUF_1
	    tmp_dev->bar0[0x2700+i+32*j] = E1_TO_CTB_TABLE(1,(i*4+j));
	}
    }
    i = 0x76; //6CT=> CHAN0 7CT=>CHAN1
    tmp_dev->bar0[0x08] = i;
    /*Mapping in DMA buffers is the same as in case of acess point first come 0 framer and 0 time slot*/
}

static void init_framer_regs(const struct pmc_e1_controller * tmp_dev)
{

    int addr=0x800; //0x200 -> 0x800  new firmware!
    int j;
    if (PMC_OLD_FIRMWARE == (tmp_dev->bar0[0x0D] & 0xffff))
    {
	    tmp_dev->bar0[0x03] = 0x00;  //reset all framers
	    udelay(1000); //wait
	    tmp_dev->bar0[0x03] = 0xff;
	    udelay(1000); //wait
    }

    tmp_dev->bar1[addr] = 0x02;  // setup page 2
    tmp_dev->bar1[addr + 0x10] = 0xa0; //Configuration control word: E1 mode, Transmit enable
    tmp_dev->bar1[addr + 0x11] = 0; //P2.11=0h (Pulse attenuation) for 120 Om line. LIU control word
    tmp_dev->bar1[addr + 0x12] = 0; //write 0's to reserved space
    tmp_dev->bar1[addr + 0x14] = 0;
    tmp_dev->bar1[addr + 0x15] = 0;
    tmp_dev->bar1[addr + 0x16] = 0xbb; //P2.16 eqalizer high threshold
    tmp_dev->bar1[addr + 0x17] = 0x30; //eqalizer low threshold

    //setup page 1
    tmp_dev->bar1[addr] = 0x01;
    tmp_dev->bar1[addr + 0x14] = 0x80; //to output data to dsto, select CAS;
    tmp_dev->bar1[addr + 0x19] = 0x01; //to obtain loss signal
    tmp_dev->bar1[addr + 0x1f] = 0x18; //to setup automatic equaliser
    tmp_dev->bar1[addr + 0x15] = 0xa0; //P1.15 Coding and loopback controll HDB both directions;
    tmp_dev->bar1[addr + 0x13] = 0x0b; //setup defaults for x bits? X123=111b;
    tmp_dev->bar1[addr + 0x12] = 0x03; //setup defaults for International bits I01=11b
    tmp_dev->bar1[addr + 0x11] = 0x10;  //disable TxAO

    //cas related
    tmp_dev->bar1[addr + 0x1e] = 0x1;  //enable CAS interrupt

    //Enable to transmit CAS from registers on 5&6 pages.
    //setup page 7
    tmp_dev->bar1[addr] = 0x07;
   
    for(j=0;j<16;j++)
    {
	    tmp_dev->bar1[addr + 0x10 + j] = 0x02;

    }
    tmp_dev->bar1[addr] = 0x08;
    for(j=0;j<16;j++)
    {
	    tmp_dev->bar1[addr + 0x10 + j] = 0x02;
    }
    //fill 15ts CAS DATA nible with something to avoid starange initial shiftion of CAS, looks like framer hardware bug..
    tmp_dev->bar1[addr] = 0x5;
    tmp_dev->bar1[addr + 0x1f] = 0x5;
}

/*========== IOCTL (ecfg) ==========*/
/*
 static void
 check_bit(..)

 struct pmc_e1_controller * tmp_dev - pointer to pmc device to operate with
 unsigned char chan_num - number of chanal to operate
 unsigned char page - number of page of framer registers
 unsigned char line - register to operate on
 unsigned char smask - bit position in configuration uchar.
 unsigned char hmask - bit position in hardware register.
 unsigned char conf - value for comparison from pmc_config_chan structure.
 unsigned char * from_hw - value obtained from hardware after applying changes.
*/

static void check_bit(struct pmc_e1_controller * tmp_dev,
	unsigned char chan_num, unsigned char page, unsigned char line,
	unsigned char smask, const unsigned char hmask,
	unsigned char conf, unsigned char * from_hw)
{
    unsigned int tmp;
    unsigned char sstore, htore, save_smask;
    tmp = framer_read(tmp_dev, chan_num, page, line);
    htore =  (unsigned char) (tmp & hmask);  //prepare for comparison curent value from hw
    save_smask = smask;
    sstore = conf & smask; //prepare for comparison value from config
	if (smask > hmask)	//allign software value to hardware position.
	{
		while (smask > hmask)
		{
			smask>>=1;
			sstore>>= 1;
		}
	}
	else if (smask < hmask)
	{
		while (smask < hmask)
		{
			smask<<=1;
			sstore<<=1;
		}
	}
	if (htore ^ sstore) //here sstore shifted to hardware position content of configuratuon value
	{
	    tmp = framer_read(tmp_dev, chan_num, page, line);
	    tmp &= ~hmask;
	    tmp|= (sstore & hmask);
	    framer_write(tmp_dev, chan_num, page, line,tmp);
	}
        //reading back from hardware
	tmp = framer_read(tmp_dev, chan_num, page, line);
        sstore = (unsigned char) (tmp & hmask);
	smask = save_smask;
	if (smask > hmask)	//allign hardware to software position.
	{
	    while (smask > hmask)
	    {
		smask>>=1;
		sstore<<=1;
	    }
	}
	else if (smask < hmask)
	{
	    while (smask < hmask)
	    {
		smask<<=1;
		sstore>>=1;
	    }
	}
	if (from_hw){
		*from_hw|= sstore;
	}
}

/*
Only allign to bit position in interface structure.
*/
static void align_to_swposition(const unsigned int tmp, unsigned char smask,
	const unsigned char hmask, unsigned char * aligned)
{
    unsigned char sstore;
    sstore = (unsigned char) (tmp & hmask);
    if (smask > hmask)	//allign hardware to software position.
    {
	while (smask > hmask)
	{
	    smask>>=1;
	    sstore<<=1;
	}
    }
    else if (smask < hmask)
    {
	while (smask < hmask)
	{
	    smask<<=1;
	    sstore>>=1;
	}
    }
    if (aligned){
	*aligned|= sstore;
    }
}

/*
 Return codes:
 ENOTTY - invalid ioctl;
 EINVAL - passing invalid parameters;
 EFAULT - error copying data from userspace;
 0 - success;
*/
static int pmc_e1_ioctl(struct inode * inode_p, struct file * file_p, unsigned int cmd, unsigned long arg)
{
	int err = 0, k, j;
	unsigned int tmp=0, channo = 0;
	unsigned char ucTmp=0;
	struct pmc_config_chan store_chanl_conf, read_back_chan_conf;
	struct pmc_err_ins err_ins;
	struct pmc_data_pass pass_data;
	struct pmc_status_chan stat_exchange;
	struct pmc_config_board board_config;
	struct pmc_cas_xmit_data cas_data;
	struct pmc_version version_exchange;
	struct pmc_e1_controller * tmp_dev = (struct pmc_e1_controller *) file_p->private_data;

	if (_IOC_TYPE(cmd) != PTMC_E1_8_MAGIC)
		return -ENOTTY;
	if (_IOC_NR(cmd) > PTMC_E1_8_MCN)
		return -ENOTTY;
	if (_IOC_DIR(cmd) & _IOC_READ)
		err = !access_ok(VERIFY_WRITE, (void *) arg, _IOC_SIZE(cmd));
	else if (_IOC_DIR(cmd) & _IOC_WRITE)
		err = !access_ok(VERIFY_READ, (void *) arg, _IOC_SIZE(cmd));
	if (err)
		return -EFAULT;

	memset((void *) &board_config, 0, sizeof(struct pmc_config_board));
	memset((void *) &read_back_chan_conf, 0, sizeof(struct pmc_config_chan));
	memset((void *) &store_chanl_conf, 0, sizeof(struct pmc_config_chan));
	memset((void *) &err_ins, 0, sizeof(struct pmc_err_ins));
	memset((void *) &stat_exchange, 0, sizeof(struct pmc_status_chan));
	memset((void *) &cas_data, 0, sizeof(struct pmc_cas_xmit_data));

	switch(cmd) {
	case PTMC_E1_CHAN_CONFIGURE:
		module_dprintk(KERN_DEBUG, "PTMC_E1_CHAN_CONFIGURE.\n");
		if(copy_from_user((void *) &store_chanl_conf, (void *) arg, sizeof(store_chanl_conf))) {
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
                channo = store_chanl_conf.chan_num;
		if ((channo & (~DIRECT_IN)) > tmp_dev->e1_amount)
			return -EINVAL;

		/* Write port configuration*/
		if (!(channo & DIRECT_IN)) {
			channo &= (~DIRECT_IN);
			//CFG0
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG0_CAS_HP),  PMC_GREG(CFG0_CAS_HP), CFG0_CAS, PMC_GBIT(CFG0_CAS_HP), store_chanl_conf.cfg0,  &read_back_chan_conf.cfg0);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG0_RCV_TR_HP), PMC_GREG(CFG0_RCV_TR_HP), CFG0_RCV_TR, PMC_GBIT(CFG0_RCV_TR_HP), store_chanl_conf.cfg0,  &read_back_chan_conf.cfg0);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG0_XMIT_TR_HP), PMC_GREG(CFG0_XMIT_TR_HP), CFG0_XMIT_TR, PMC_GBIT(CFG0_XMIT_TR_HP), store_chanl_conf.cfg0,  &read_back_chan_conf.cfg0);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG0_RCV_HDB3_HP), PMC_GREG(CFG0_RCV_HDB3_HP), CFG0_RCV_HDB3, PMC_GBIT(CFG0_RCV_HDB3_HP), store_chanl_conf.cfg0,  &read_back_chan_conf.cfg0);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG0_XMIT_HDB3_HP), PMC_GREG(CFG0_XMIT_HDB3_HP), CFG0_XMIT_HDB3, PMC_GBIT(CFG0_XMIT_HDB3_HP), store_chanl_conf.cfg0,  &read_back_chan_conf.cfg0);
			//CFG1
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG1_NAT0_HP), PMC_GREG(CFG1_NAT0_HP), CFG1_NAT0,  PMC_GBIT(CFG1_NAT0_HP), store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG1_NAT1_HP), PMC_GREG(CFG1_NAT1_HP), CFG1_NAT1,  PMC_GBIT(CFG1_NAT1_HP), store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG1_NAT2_HP), PMC_GREG(CFG1_NAT2_HP), CFG1_NAT2,  PMC_GBIT(CFG1_NAT2_HP), store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG1_NAT3_HP), PMC_GREG(CFG1_NAT3_HP), CFG1_NAT3,  PMC_GBIT(CFG1_NAT3_HP), store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG1_NAT4_HP), PMC_GREG(CFG1_NAT4_HP), CFG1_NAT4,  PMC_GBIT(CFG1_NAT4_HP), store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG1_INAT0_HP), PMC_GREG(CFG1_INAT0_HP), CFG1_INAT0,  PMC_GBIT(CFG1_INAT0_HP), store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG1_INAT1_HP), PMC_GREG(CFG1_INAT1_HP), CFG1_INAT1,  PMC_GBIT(CFG1_INAT1_HP), store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			//CFG2
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG2_RAI_AUTO_ON_HP), PMC_GREG(CFG2_RAI_AUTO_ON_HP), CFG2_RAI_MANUAL, PMC_GBIT(CFG2_RAI_AUTO_ON_HP), store_chanl_conf.cfg2,  &read_back_chan_conf.cfg2);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG2_RAI_ENA_HP), PMC_GREG(CFG2_RAI_ENA_HP), CFG2_RAI_ON,  PMC_GBIT(CFG2_RAI_ENA_HP), store_chanl_conf.cfg2,  &read_back_chan_conf.cfg2);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG2_RCAS_AUTO_ON_HP), PMC_GREG(CFG2_RCAS_AUTO_ON_HP), CFG2_RCAS_MANUAL,  PMC_GBIT(CFG2_RCAS_AUTO_ON_HP), store_chanl_conf.cfg2,  &read_back_chan_conf.cfg2);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG2_RCAS_ENA_HP), PMC_GREG(CFG2_RCAS_ENA_HP), CFG2_RCAS_ON,  PMC_GBIT(CFG2_RCAS_ENA_HP), store_chanl_conf.cfg2,  &read_back_chan_conf.cfg2);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG2_X1_HP), PMC_GREG(CFG2_X1_HP), CFG2_X1, PMC_GBIT(CFG2_X1_HP), store_chanl_conf.cfg2,  &read_back_chan_conf.cfg2);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG2_X2_HP), PMC_GREG(CFG2_X2_HP), CFG2_X2, PMC_GBIT(CFG2_X2_HP), store_chanl_conf.cfg2,  &read_back_chan_conf.cfg2);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG2_X3_HP), PMC_GREG(CFG2_X3_HP), CFG2_X3, PMC_GBIT(CFG2_X3_HP), store_chanl_conf.cfg2,  &read_back_chan_conf.cfg2);
			//CFG3
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG3_CRC4_TR_HP), PMC_GREG(CFG3_CRC4_TR_HP), CFG3_CRC4_TR,  PMC_GBIT(CFG3_CRC4_TR_HP), store_chanl_conf.cfg3,  &read_back_chan_conf.cfg3);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG3_CRC4_DIS_HP), PMC_GREG(CFG3_CRC4_DIS_HP), CFG3_CRC4_DIS,  PMC_GBIT(CFG3_CRC4_DIS_HP), store_chanl_conf.cfg3,  &read_back_chan_conf.cfg3);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG3_CRC4_MANUAL_HP), PMC_GREG(CFG3_CRC4_MANUAL_HP), CFG3_CRC4_MANUAL,  PMC_GBIT(CFG3_CRC4_MANUAL_HP), store_chanl_conf.cfg3,  &read_back_chan_conf.cfg3);
			//CFG4
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG4_AIS_HP), PMC_GREG(CFG4_AIS_HP), CFG4_AIS,  PMC_GBIT(CFG4_AIS_HP), store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG4_LLOOP_HP), PMC_GREG(CFG4_LLOOP_HP), CFG4_LLOOP,  PMC_GBIT(CFG4_LLOOP_HP), store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG4_RLOOP_HP), PMC_GREG(CFG4_RLOOP_HP), CFG4_RLOOP,  PMC_GBIT(CFG4_RLOOP_HP), store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);
			check_bit(tmp_dev, channo, PMC_GPAGE(CFG4_PLOOP_HP), PMC_GREG(CFG4_PLOOP_HP), CFG4_PLOOP,  PMC_GBIT(CFG4_PLOOP_HP), store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);

			/* configure board part */
			if (store_chanl_conf.cfg0 & CFG0_MASTER) {
				tmp = tmp_dev->bar0[0x06]; //get bs/ls
				ucTmp = (1<<channo);
				tmp|= (unsigned int) ucTmp;
				tmp_dev->bar0[0x06] = tmp; //restore bs/ls
				read_back_chan_conf.cfg0|= CFG0_MASTER;
			} else { /*allow to sync to this chan */
				tmp = tmp_dev->bar0[0x06]; //get bs/ls
				ucTmp = ~(1<<channo);
				tmp&= (unsigned int) ucTmp;
				tmp_dev->bar0[0x06] = tmp; //restore bs/ls
			}
			if (store_chanl_conf.cfg4 & CFG4_TX_DIS) {
				/* disable transmition, will transmit all ones */
				tmp = tmp_dev->bar0[0x04]; //get TXAO
				ucTmp = ~(1<<channo);
				tmp&= ucTmp;
				tmp_dev->bar0[0x04] = tmp; //restore TXAO
				read_back_chan_conf.cfg4|= CFG4_TX_DIS;
			} else { /* enable transmition */
				tmp = tmp_dev->bar0[0x04]; //get TXAO
				ucTmp = (1<<channo);
				tmp|= ucTmp;
				tmp_dev->bar0[0x04] = tmp; //restore TXAO
			}
		} else { /* Read port configuration */
			channo &= (~DIRECT_IN);
			//1-10
			tmp = framer_read(tmp_dev, channo, 0x1, 0x10);
			align_to_swposition(tmp, CFG3_CRC4_DIS, PMC_GBIT(CFG3_CRC4_DIS_HP), &read_back_chan_conf.cfg3); //1-10-2
			align_to_swposition(tmp, CFG2_RCAS_MANUAL, PMC_GBIT(CFG2_RCAS_AUTO_ON_HP), &read_back_chan_conf.cfg2); //1-10-3
			align_to_swposition(tmp, CFG2_RAI_MANUAL, PMC_GBIT(CFG2_RAI_AUTO_ON_HP), &read_back_chan_conf.cfg2); //1-10-4
			align_to_swposition(tmp, CFG3_CRC4_MANUAL, PMC_GBIT(CFG3_CRC4_MANUAL_HP), &read_back_chan_conf.cfg3);//1-10-5
			align_to_swposition(tmp, CFG3_CRC4_TR, PMC_GBIT(CFG3_CRC4_TR_HP), &read_back_chan_conf.cfg3 ); //1-10-6

			//1-11
			tmp = framer_read(tmp_dev, channo, 0x1, 0x11);
			align_to_swposition(tmp, CFG4_AIS, PMC_GBIT(CFG4_AIS_HP), &read_back_chan_conf.cfg4); //1-11-4
			//1-12
			tmp = framer_read(tmp_dev, channo, 0x1, 0x12);
			align_to_swposition(tmp, CFG1_INAT0, PMC_GBIT(CFG1_INAT0_HP), &read_back_chan_conf.cfg1); //1-12-0
			align_to_swposition(tmp, CFG1_INAT1, PMC_GBIT(CFG1_INAT1_HP), &read_back_chan_conf.cfg1); //1-12-1
			align_to_swposition(tmp, CFG0_XMIT_TR, PMC_GBIT(CFG0_XMIT_TR_HP), &read_back_chan_conf.cfg0); //1-12-2
			align_to_swposition(tmp, CFG0_RCV_TR, PMC_GBIT(CFG0_RCV_TR_HP), &read_back_chan_conf.cfg0); //1-12-3
			//1-13
			tmp = framer_read(tmp_dev, channo, 0x1, 0x13);
			align_to_swposition(tmp, CFG2_X3, PMC_GBIT(CFG2_X3_HP), &read_back_chan_conf.cfg2); //1-13-0
			align_to_swposition(tmp, CFG2_X2, PMC_GBIT(CFG2_X2_HP), &read_back_chan_conf.cfg2); //1-13-1
			align_to_swposition(tmp, CFG2_RCAS_ON, PMC_GBIT(CFG2_RCAS_ENA_HP), &read_back_chan_conf.cfg2 ); //1-13-2
			align_to_swposition(tmp, CFG2_X1, PMC_GBIT(CFG2_X1_HP), &read_back_chan_conf.cfg2 ); //1-13-3

			//1-14
			tmp = framer_read(tmp_dev, channo, 0x1, 0x14);
			align_to_swposition(tmp, CFG0_CAS, PMC_GBIT(CFG0_CAS_HP), &read_back_chan_conf.cfg0); //1-14-5
			//1-15
			tmp = framer_read(tmp_dev, channo, 0x1, 0x15);
			align_to_swposition(tmp, CFG4_PLOOP, PMC_GBIT(CFG4_PLOOP_HP), &read_back_chan_conf.cfg4); //1-15-0
			align_to_swposition(tmp, CFG4_RLOOP, PMC_GBIT(CFG4_RLOOP_HP), &read_back_chan_conf.cfg4); //1-15-2
			align_to_swposition(tmp, CFG0_XMIT_HDB3, PMC_GBIT(CFG0_XMIT_HDB3_HP), &read_back_chan_conf.cfg0); //1-15-5
			align_to_swposition(tmp, CFG4_LLOOP, PMC_GBIT(CFG4_LLOOP_HP), &read_back_chan_conf.cfg4); //1-15-6
			align_to_swposition(tmp, CFG0_RCV_HDB3, PMC_GBIT(CFG0_RCV_HDB3_HP), &read_back_chan_conf.cfg0); //1-15-7
			//1-16
			tmp = framer_read(tmp_dev, channo, 0x1, 0x16);
			align_to_swposition(tmp, CFG2_RAI_ON, PMC_GBIT(CFG2_RAI_ENA_HP), &read_back_chan_conf.cfg2); //1-16-5
			//1-17
			tmp = framer_read(tmp_dev, channo, 0x1, 0x17);
			align_to_swposition(tmp, CFG1_NAT4, PMC_GBIT(CFG1_NAT4_HP), &read_back_chan_conf.cfg1); //1-17-0
			align_to_swposition(tmp, CFG1_NAT3, PMC_GBIT(CFG1_NAT3_HP), &read_back_chan_conf.cfg1); //1-17-1
			align_to_swposition(tmp, CFG1_NAT2, PMC_GBIT(CFG1_NAT2_HP), &read_back_chan_conf.cfg1); //1-17-2
			align_to_swposition(tmp, CFG1_NAT1, PMC_GBIT(CFG1_NAT1_HP), &read_back_chan_conf.cfg1); //1-17-3
			align_to_swposition(tmp, CFG1_NAT0, PMC_GBIT(CFG1_NAT0_HP), &read_back_chan_conf.cfg1); //1-17-4

			/* configure board part */
			/* get BS/LS */
			tmp = (unsigned char) tmp_dev->bar0[0x06];
			ucTmp = (1<<channo);
			align_to_swposition(tmp, CFG0_MASTER, ucTmp, &read_back_chan_conf.cfg0); //0-6-chan_dep


			/* get TXAO */
			tmp = ~(unsigned char) (tmp_dev->bar0[0x04]);
			ucTmp = (1<<channo);
			align_to_swposition(tmp, CFG4_TX_DIS, ucTmp, &read_back_chan_conf.cfg4); //0-4-chan_dep
		}
		if(copy_to_user((void *) arg, (void *) &read_back_chan_conf, sizeof(read_back_chan_conf))) {
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		break;
	case PTMC_E1_ERR_EMU:
		module_dprintk(KERN_DEBUG, "PTMC_E1_ERR_EMU.\n");

		if (copy_from_user((void *) &err_ins, (void *) arg, sizeof(err_ins))) {
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
                channo = err_ins.chan_num;
		if (channo > tmp_dev->e1_amount)
			return -EINVAL;

		//self restore zero state;
		check_bit(tmp_dev, channo, PMC_GPAGE(ERR_EMU0_LOF_NFAS_HP),  PMC_GREG(ERR_EMU0_LOF_NFAS_HP), ERR_EMU0_LOF_NFAS , PMC_GBIT(ERR_EMU0_LOF_NFAS_HP), 0, NULL);
		check_bit(tmp_dev, channo, PMC_GPAGE(ERR_EMU0_LOF_FAS_HP),  PMC_GREG(ERR_EMU0_LOF_FAS_HP), ERR_EMU0_LOF_FAS , PMC_GBIT(ERR_EMU0_LOF_FAS_HP), 0, NULL);
		check_bit(tmp_dev, channo, PMC_GPAGE(ERR_EMU0_CRC4_HP),  PMC_GREG(ERR_EMU0_CRC4_HP), ERR_EMU0_CRC4, PMC_GBIT(ERR_EMU0_CRC4_HP), 0, NULL);
		//generate errors
		check_bit(tmp_dev, channo, PMC_GPAGE(ERR_EMU0_LOF_NFAS_HP),  PMC_GREG(ERR_EMU0_LOF_NFAS_HP), ERR_EMU0_LOF_NFAS , PMC_GBIT(ERR_EMU0_LOF_NFAS_HP), err_ins.pmc_err, NULL);
		check_bit(tmp_dev, channo, PMC_GPAGE(ERR_EMU0_LOF_FAS_HP),  PMC_GREG(ERR_EMU0_LOF_FAS_HP), ERR_EMU0_LOF_FAS , PMC_GBIT(ERR_EMU0_LOF_FAS_HP), err_ins.pmc_err, NULL);
		check_bit(tmp_dev, channo, PMC_GPAGE(ERR_EMU0_CRC4_HP),  PMC_GREG(ERR_EMU0_CRC4_HP), ERR_EMU0_CRC4, PMC_GBIT(ERR_EMU0_CRC4_HP), err_ins.pmc_err, NULL);
		//self restore zero state;
		check_bit(tmp_dev, channo, PMC_GPAGE(ERR_EMU0_LOF_NFAS_HP),  PMC_GREG(ERR_EMU0_LOF_NFAS_HP), ERR_EMU0_LOF_NFAS , PMC_GBIT(ERR_EMU0_LOF_NFAS_HP), 0, NULL);
		check_bit(tmp_dev, channo, PMC_GPAGE(ERR_EMU0_LOF_FAS_HP),  PMC_GREG(ERR_EMU0_LOF_FAS_HP), ERR_EMU0_LOF_FAS , PMC_GBIT(ERR_EMU0_LOF_FAS_HP), 0, NULL);
		check_bit(tmp_dev, channo, PMC_GPAGE(ERR_EMU0_CRC4_HP),  PMC_GREG(ERR_EMU0_CRC4_HP), ERR_EMU0_CRC4, PMC_GBIT(ERR_EMU0_CRC4_HP), 0, NULL);

		break;
	case PTMC_E1_GET_STATS:
		module_dprintk(KERN_DEBUG, "PTMC_E1_GET_STATS.\n");

		if (copy_from_user((void *) &stat_exchange, (void *) arg, sizeof(stat_exchange))) {
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
                channo = stat_exchange.chan_num & (~DIRECT_IN);
		if (channo > tmp_dev->e1_amount)
			return -EINVAL;

		//3-10
		tmp = framer_read(tmp_dev, channo, 0x3, 0x10);
		align_to_swposition(tmp, STAT0_LOF, PMC_GBIT(STAT0_LOF_HP), &stat_exchange.stat0);
		align_to_swposition(tmp, STAT0_CAS_LOM, PMC_GBIT(STAT0_CAS_LOM_HP), &stat_exchange.stat0);
		align_to_swposition(tmp, STAT0_CRC4_SYNC, PMC_GBIT(STAT0_CRC4_SYNC_HP), &stat_exchange.stat0);
		if (!(  (tmp & PMC_GBIT(STAT0_E_BIT1_HP)) && (tmp & PMC_GBIT(STAT0_E_BIT2_HP)) )) {
			stat_exchange.stat0 |= STAT0_E_BIT;
		}

		//3-11
		tmp = framer_read(tmp_dev, channo, 0x3, 0x11);
		align_to_swposition(tmp, STAT0_LOS, PMC_GBIT(STAT0_LOS_HP), &stat_exchange.stat0);
		align_to_swposition(tmp, STAT0_AIS, PMC_GBIT(STAT0_AIS_HP), &stat_exchange.stat0);
		align_to_swposition(tmp, STAT0_RAIS, PMC_GBIT(STAT0_RAIS_HP), &stat_exchange.stat0);
		if ((tmp & PMC_GBIT(STAT0_CRC4_1_HP)) || (tmp & PMC_GBIT(STAT0_CRC4_2_HP))) {
			stat_exchange.stat0 |= STAT0_CRC4;
		}

		//3-15
		tmp = framer_read(tmp_dev, channo, 0x3, 0x15);
		align_to_swposition(tmp, STAT1_INAT0, PMC_GBIT(STAT1_INAT0_HP), &stat_exchange.stat1);

		//3-18
		tmp = framer_read(tmp_dev, channo, 0x3, 0x18);
		align_to_swposition(tmp, STAT1_NAT0, PMC_GBIT(STAT1_NAT0_HP), &stat_exchange.stat1);
		align_to_swposition(tmp, STAT1_NAT1, PMC_GBIT(STAT1_NAT1_HP), &stat_exchange.stat1);
		align_to_swposition(tmp, STAT1_NAT2, PMC_GBIT(STAT1_NAT2_HP), &stat_exchange.stat1);
		align_to_swposition(tmp, STAT1_NAT3, PMC_GBIT(STAT1_NAT3_HP), &stat_exchange.stat1);
		align_to_swposition(tmp, STAT1_NAT4, PMC_GBIT(STAT1_NAT4_HP), &stat_exchange.stat1);
		align_to_swposition(tmp, STAT1_INAT1, PMC_GBIT(STAT1_INAT1_HP), &stat_exchange.stat1);

		//3-19
		tmp = framer_read(tmp_dev, channo, 0x3, 0x19);
		align_to_swposition(tmp, STAT2_X1, PMC_GBIT(STAT2_X1_HP), &stat_exchange.stat2);
		align_to_swposition(tmp, STAT2_X2, PMC_GBIT(STAT2_X2_HP), &stat_exchange.stat2);
		align_to_swposition(tmp, STAT2_X3, PMC_GBIT(STAT2_X3_HP), &stat_exchange.stat2);
		align_to_swposition(tmp, STAT2_Y, PMC_GBIT(STAT2_Y_HP), &stat_exchange.stat2);

		//4-12
		tmp = framer_read(tmp_dev, channo, 0x4, 0x12);
		align_to_swposition(tmp, STAT1_SLIP, PMC_GBIT(STAT1_SLIP_HP), &stat_exchange.stat1);


		/* reading FAS counter */
		stat_exchange.fas = framer_read_pr(tmp_dev, channo, CNT_FAS_HP);
		/* reading HDB3 counter */
		stat_exchange.hdb3 = framer_read_pr(tmp_dev, channo, CNT_BPVHI_HP);
		stat_exchange.hdb3<<=8;
		stat_exchange.hdb3 |= framer_read_pr(tmp_dev, channo, CNT_BPVLO_HP);
		/* reading CRC4 counter */
		stat_exchange.crc4 = framer_read_pr(tmp_dev, channo, CNT_CRC4HI_HP);
		stat_exchange.crc4<<=8;
		stat_exchange.crc4 |= framer_read_pr(tmp_dev, channo, CNT_CRC4LO_HP);

		/* Update overrun flags */
		//4-1f
		tmp = framer_read(tmp_dev, channo, 0x4, 0x1f);
		align_to_swposition(tmp, OVR_HDB3, PMC_GBIT(OVR_HDB3_HP), &stat_exchange.ovr);
		align_to_swposition(tmp, OVR_FAS, PMC_GBIT(OVR_FAS_HP), &stat_exchange.ovr);
		align_to_swposition(tmp, OVR_CRC4, PMC_GBIT(OVR_CRC4_HP), &stat_exchange.ovr);
		/* clear counter if requested */
		if (stat_exchange.mgmt & OVR_CRC4) {
			framer_write_pr(tmp_dev, channo, CNT_CRC4HI_HP, 0);
			framer_write_pr(tmp_dev, channo, CNT_CRC4LO_HP, 0);
		}
		if (stat_exchange.mgmt & OVR_HDB3) {
			framer_write_pr(tmp_dev, channo, CNT_BPVHI_HP, 0);
			framer_write_pr(tmp_dev, channo, CNT_BPVLO_HP, 0);
		}
		if (stat_exchange.mgmt & OVR_FAS) {
			framer_write_pr(tmp_dev, channo, CNT_FAS_HP, 0);
		}

		/* get cas signalling */
		k=0;
		for (j=0; j<16; j++) {
			tmp = framer_read(tmp_dev, channo, 0x9, 0x10 + j);
			if (j < (2*k+1)) {
				stat_exchange.cas_data[k] = (unsigned char) (tmp & 0xf);
				stat_exchange.cas_data[k] <<= 4;

			} else {
				tmp&=0x0f;
				stat_exchange.cas_data[k] |= (unsigned char) tmp;
				k+=1;
			}
		}
		k=0;
		for (j=0; j<16; j++) {
			tmp = framer_read(tmp_dev, channo, 0xA, 0x10 + j);
			if (j < (2*k+1)) {
				stat_exchange.cas_data[8+k] = (unsigned char) (tmp & 0xf);
				stat_exchange.cas_data[8+k] <<= 4;
			} else {
				tmp&=0x0f;
				stat_exchange.cas_data[8+k] |= (unsigned char) tmp;
				k+=1;
			}
		}

		if(copy_to_user((void *) arg, (void *) &stat_exchange, sizeof(stat_exchange))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		break;
	case PTMC_E1_RESET:
		/* Reset board. Not tested very well. */
		module_dprintk(KERN_DEBUG, "PTMC_E1_RESET.\n");

		/* initialising board registers */
		init_ptmc_regs(tmp_dev);
		/* initialising framer registers */
		init_framer_regs(tmp_dev);
		/* initialising time-slots maping table */
		init_switch_map(tmp_dev);
		break;
	case PTMC_E1_BOARD_CONFIGURE:
		module_dprintk(KERN_DEBUG, "PTMC_E1_BOARD_CONFIGURE.\n");

		if(copy_from_user((void *) &board_config, (void *) arg, _IOC_SIZE(cmd))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		/* reading ctrl register */
		tmp = tmp_dev->bar0[0x02];
		tmp &= ~(MODE1_NF1_GEN_ENA_HP | MODE1_NF2_GEN_ENA_HP | MODE1_NF1_USE_ENA_HP | MODE1_NF2_USE_ENA_HP | MODE1_AB_HP | MODE1_MA_SL_SEL_HP); //set to zero nessasitive bit fild.
		if (board_config.mode1 & MODE1_NF1_GEN_ENA) {
			tmp|= MODE1_NF1_GEN_ENA_HP;
		}
		if (board_config.mode1 & MODE1_NF2_GEN_ENA) {
			tmp|= MODE1_NF2_GEN_ENA_HP;
		}
		if (board_config.mode1 & MODE1_NF1_USE_ENA) {
			tmp|= MODE1_NF1_USE_ENA_HP;
		}
		if (board_config.mode1 & MODE1_NF2_USE_ENA) {
			tmp|= MODE1_NF2_USE_ENA_HP;
		}
		if (board_config.mode1 & MODE1_AB) {
			tmp|= MODE1_AB_HP;
		}
		if (board_config.mode1 & MODE1_MA_SL_SEL) {
			tmp|= MODE1_MA_SL_SEL_HP;
		}
		tmp_dev->bar0[0x02] = tmp;
		udelay(2000);
		break;
	case PTMC_E1_GET_PCI_SLOT:
		module_dprintk(KERN_DEBUG, "PTMC_E1_GET_PCI_SLOT.\n");

		tmp = PCI_SLOT( (unsigned int) tmp_dev->pci_device->devfn);
		if (copy_to_user((void *) arg, (void *) &tmp, _IOC_SIZE(cmd))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		break;
	case PTMC_E1_GET_PCI_BUSNUM:
		module_dprintk(KERN_DEBUG, "PTMC_E1_GET_PCI_BUSNUM.\n");

		tmp = tmp_dev->pci_device->bus->number;
		if (copy_to_user((void *) arg, (void *) &tmp, _IOC_SIZE(cmd))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		break;
	case PTMC_E1_READ_BAR0:
		module_dprintk(KERN_DEBUG, "PTMC_E1_READ_BAR0.\n");

		if (copy_from_user((void *) &pass_data, (void *) arg, sizeof(pass_data))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF)
			return -EINVAL;
		pass_data.rdata = tmp_dev->bar0[pass_data.raddress];
		if (copy_to_user((void *) arg, (void *) &pass_data, sizeof(pass_data))){
			return -EFAULT;
		}
		break;
	case PTMC_E1_WRITE_BAR0:
		module_dprintk(KERN_DEBUG, "PTMC_E1_WRITE_BAR0.\n");

		if (copy_from_user((void *) &pass_data, (void *) arg, sizeof(pass_data))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF)
			return -EINVAL;
		tmp_dev->bar0[pass_data.raddress] = pass_data.rdata;
		break;
	case PTMC_E1_READ_BAR1:
		module_dprintk(KERN_DEBUG, "PTMC_E1_READ_BAR1.\n");

		if (copy_from_user((void *) &pass_data, (void *) arg, sizeof(pass_data))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF)
			return -EINVAL;
		pass_data.rdata = tmp_dev->bar1[pass_data.raddress];
		if (copy_to_user((void *) arg, (void *) &pass_data, sizeof(pass_data)) ) {
			return -EFAULT;
		}
		break;
	case PTMC_E1_WRITE_BAR1:
		module_dprintk(KERN_DEBUG, "PTMC_E1_WRITE_BAR1.\n");

		if (copy_from_user((void *) &pass_data, (void *) arg, sizeof(pass_data))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF)
			return -EINVAL;
		tmp_dev->bar1[pass_data.raddress] = pass_data.rdata;
		break;
	case PTMC_E1_CAS_XMIT_PASS:
		module_dprintk(KERN_DEBUG, "PTMC_E1_CAS_XMIT_PASS\n");

		if(copy_from_user((void *) &cas_data, (void *) arg, _IOC_SIZE(cmd))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
                channo = cas_data.chan_num;
		if ((channo & (~DIRECT_IN)) > tmp_dev->e1_amount)
			return -EINVAL;

		if (!(channo & DIRECT_IN)) { /* writing cas data */
			channo &= (~DIRECT_IN);
			module_dprintk(KERN_DEBUG, "Pass data! Chan_num = %d\n", channo);

			k=0;
			for (j=0; j<16; j++) {
				if (j < (2*k+1)) {
					tmp = (cas_data.cdata[k] >> 4) & 0x0f;
					framer_write(tmp_dev, channo, 0x5, 0x10 + j, tmp);
				} else {
					tmp = (cas_data.cdata[k] & 0x0f);
					framer_write(tmp_dev, channo, 0x5, 0x10 + j, tmp);
					k+=1;
				}
			}
			k=0;
			for (j=0; j<16; j++) {
				if (j < (2*k+1)) {
					tmp = (cas_data.cdata[8+k] >> 4) & 0x0f;
					framer_write(tmp_dev, channo, 0x6, 0x10 + j, tmp);
				} else {
					tmp = (cas_data.cdata[8+k] & 0x0f);
					framer_write(tmp_dev, channo, 0x6, 0x10 + j, tmp);
					k+=1;
				}
			}
			memset((void *) &cas_data, 0, sizeof(struct pmc_cas_xmit_data)); //clear to get values back
		}
		/* Reading back cas signalling */
		channo &= (~DIRECT_IN);
		k=0;
		for (j=0; j<16; j++) {
			tmp = framer_read(tmp_dev, channo, 0x5, 0x10 + j);
			if (j < (2*k+1)) {
				tmp&=0x0f;
				cas_data.cdata[k] = (unsigned char) tmp;
				cas_data.cdata[k] <<= 4;
			} else {
				tmp&=0x0f;
				cas_data.cdata[k] |= (unsigned char) tmp;
				k+=1;
			}

		}
		k=0;
		for (j=0; j<16; j++) {
			tmp = framer_read(tmp_dev, channo, 0x6, 0x10 + j);
			if (j < (2*k+1)) {
				tmp&=0x0f;
				cas_data.cdata[8+k] = (unsigned char) tmp;
				cas_data.cdata[8+k] <<= 4;
			} else {
				tmp&=0x0f;
				cas_data.cdata[8+k] |= (unsigned char) tmp;
				k+=1;
			}
		}

		if (copy_to_user((void *) arg, (void *) &cas_data, sizeof(cas_data))) {
			return -EFAULT;
		}
		break;
	case PTMC_E1_GET_VERSION:
		module_dprintk(KERN_DEBUG, "PTMC_E1_VERSION.\n");
		strncpy (version_exchange.proto_version, ecfg_proto_version, sizeof (version_exchange.proto_version));
		version_exchange.hw_fw_rev = tmp_dev->bar0[0x0D];
		strncpy (version_exchange.driver_version, QUASAR_VERSION, sizeof (version_exchange.driver_version));
		if(copy_to_user((void *) arg, (void *) &version_exchange, sizeof(version_exchange))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		break;
		/*
		 case PTMC_E1_PASS_ABCD_ZAP_FAKE:
		 //printk("PTMC_E1_PASS_ABCD_ZAP_FAKE\n");
		 if(copy_from_user((void *) &tmp, (void *) arg, _IOC_SIZE(cmd)))
		 {
		 printk(KERN_ERR DRV_NAME "Memory access fault!\n");
		 return -EFAULT;
		 }
		 tmp_dev->test_sig = tmp;
		 //printk("Passing:0x%x\n", tmp);
		 break;
		 */
	default:
		break;
	}
	return 0;
}

#ifdef HAVE_UNLOCKED_IOCTL
static long pmc_e1_ioctl_unlocked(struct file *file, unsigned int cmd, unsigned long data)
{
	long res;

	lock_kernel();
	res = pmc_e1_ioctl(NULL, file, cmd, data);
	unlock_kernel();
        return res;
}
#endif

static int pmc_e1_open(struct inode * inode_p, struct file * file_p)
{
	file_p->private_data = container_of(inode_p->i_cdev, struct pmc_e1_controller, dev_file);
	return 0;
}

static int pmc_e1_release(struct inode * inode_p, struct file * file_p)
{
	return 0;
}

static struct file_operations pmc_e1_fops  = {
	owner:		THIS_MODULE,
	open:		pmc_e1_open,
#ifdef HAVE_UNLOCKED_IOCTL
	unlocked_ioctl:	pmc_e1_ioctl_unlocked,
#else
	ioctl:		pmc_e1_ioctl,
#endif
    release:            pmc_e1_release,
};

/*========== SYSFS code ==========*/
#ifdef PMC_E1_UDEV
struct ptmc_attribute {
    struct device_attribute attr;
    unsigned long p[3];
};
#define toptattr(a) ((struct ptmc_attribute *)(a))
#define todeattr(a) ((struct device_attribute *)(a))

#define PTMC_ATTR_RO(name) __ATTR(name, S_IRUGO, _pmc_show, NULL)
ssize_t _pmc_show(struct device *cls, struct device_attribute *attr,
		      char *buff)
{
    struct pmc_e1_controller * tmp_dev;
    unsigned long *p = toptattr(attr)->p;

    tmp_dev = (struct pmc_e1_controller *) dev_get_drvdata(cls);
    return sprintf(buff,"%d\n", (tmp_dev->bar0[p[0]] >> p[1]) && p[2]);
}

ssize_t pmc_show(struct device *cls, struct device_attribute *attr,
		 char *buff)
{
    int count=0,i,val;
    unsigned int tmp;
    struct pmc_e1_controller * tmp_dev;
    tmp_dev = (struct pmc_e1_controller *) dev_get_drvdata(cls);

    count+=sprintf(buff,"ChN:");
    for(i=0; i <= tmp_dev->e1_amount; i++)
	count+=sprintf(buff+count," %d",i);

    count+=sprintf(buff+count,"\nLOS:");
    tmp=tmp_dev->bar0[0x05];
    for(i=0; i<=tmp_dev->e1_amount; i++) {
	if (tmp & (1 << i)) val=1; else val=0;
	count+=sprintf(buff+count," %d", val);
    }

    count+=sprintf(buff+count,"\nLOF:");
    for(i=0; i<= tmp_dev->e1_amount; i++)
    {
	if (tmp & framer_checkbit(tmp_dev, i, STAT0_LOF_HP))
		val=1;
	else
		val=0;
	count+=sprintf(buff+count," %d", val);
    }
    count+=sprintf(buff+count,"\nLOM:");
    for(i=0; i <= tmp_dev->e1_amount;i++)
    {
	if (tmp & framer_checkbit(tmp_dev, i, STAT0_CAS_LOM_HP))
		val=1;
	else
		val=0;
	count+=sprintf(buff+count," %d", val);
    }

    count+=sprintf(buff+count,"\n");
    return count;
}

static struct ptmc_attribute device_attrs[] = {
    {PTMC_ATTR_RO(pmc_load), 	{0x21,16,0xFF}},
    {PTMC_ATTR_RO(pmc_load_max),{0x21,0,0xFF}},
    {PTMC_ATTR_RO(pci_load), 	{0x20,16,0xFF}},
    {PTMC_ATTR_RO(pci_load_cur),{0x20,0,0xFF}},
    {PTMC_ATTR_RO(dma_overflow),{0x09,24,0xFF}},
    {__ATTR(info, S_IRUGO, pmc_show, NULL), {0,}},
};

static int ptmc_sysfs_create (struct pmc_e1_controller * tmp_dev) {
    int i, error = 0;
    struct device *ptmc_device = tmp_dev->pmc_class_dev;

    tmp_dev->pmc_class_dev = ptmc_device =
        QUASAR_DEVICE_CREATE(pmc_e1_class, QUASAR_PARENT(tmp_dev),
		       MKDEV(MAJOR(dev_file_glob), tmp_dev->dev_number),
		       tmp_dev->name);
    if (IS_ERR(tmp_dev->pmc_class_dev)) {
        module_printk (KERN_WARNING, "Failed to device_create.\n");
	return 1;
    }
    dev_set_drvdata(tmp_dev->pmc_class_dev, (void *) tmp_dev);

    for (i = 0; i < ARRAY_SIZE(device_attrs); i++) {
	error = device_create_file(ptmc_device, &(device_attrs[i].attr));
	if (error)
	    break;
    }

    if (error) {
	while (--i >= 0)
	    device_remove_file(ptmc_device, todeattr(&device_attrs[i]));
    }

    return 0;
}

static void ptmc_sysfs_delete (struct pmc_e1_controller * tmp_dev) {
    int i;
    struct device *ptmc_device = tmp_dev->pmc_class_dev;

    if(IS_ERR(tmp_dev->pmc_class_dev)) return;

    for (i = 0; i < ARRAY_SIZE(device_attrs); i++)
	device_remove_file(ptmc_device, todeattr(&device_attrs[i]));

    device_destroy(pmc_e1_class, MKDEV(MAJOR(dev_file_glob), tmp_dev->dev_number));
}
#endif //PMC_E1_UDEV

/*========== Init code & IRQ handler ==========*/
QUASAR_IRQ_HANDLER(ptmc_e1_interrupt_heandler)
{
	unsigned int flags=0;
	struct pmc_e1_controller * tmp_dev = (struct pmc_e1_controller *) pr_data;
	if(tmp_dev == NULL ) return IRQ_NONE;
	flags = tmp_dev->bar0[0];
	if (flags & 0x20000000) //check DMA interrupt;
	{
		tmp_dev->bar0[0x0]=0x0;
		e12zap_delivery(tmp_dev);
		return IRQ_HANDLED;
	}
	return IRQ_NONE;
}

static int set_cacheline (struct pci_dev *pdev, u8 val) {
	u8 cache_tmp;

	/* Write the correct value. */
	pci_write_config_byte(pdev, PCI_CACHE_LINE_SIZE, val);
	/* Read it back. */
	pci_read_config_byte(pdev, PCI_CACHE_LINE_SIZE, &cache_tmp);
	if (val != cache_tmp) {
	    module_printk(KERN_ERR, "Cache line size of %d DWORDs is not "
			  "supported (%x:%x)\n", val,
			  pdev->vendor, pdev->device);
	    return -EINVAL;
	}

        return 0;
}

/*
 Some BIOS or Linux, set PCIe bridge "cache line size" to illegal value (>20).
 Setting "cache line size" for our FPGA & PCIe bridge (PEX 8111, PEX 8112).
 Is PCIe bridge the only device on the PCI bus (can I use first device) ?
*/
int tune_pci (struct pci_dev *pdev) {
	struct pci_bus *bus;
        struct pci_dev *bridge;
	struct list_head *next;
	u8 lat_timer, tmp_u8;

	bus = pdev->bus->parent;
        next = bus->devices.next;
	bridge = pci_dev_b (next);

 	module_printk (KERN_INFO, "Setting Cache Line Size to %d\n", cacheline);
	if (set_cacheline (pdev, cacheline))
            return -EINVAL;

	pci_read_config_byte(pdev, PCI_LATENCY_TIMER, &lat_timer);
	if (lat_timer < 40 || latency) {
		lat_timer = latency ? latency : 40;
		module_printk (KERN_INFO, "Setting latency timer to %d\n",
			       lat_timer);
		pci_write_config_byte(pdev, PCI_LATENCY_TIMER, lat_timer);
		pci_read_config_byte(pdev, PCI_LATENCY_TIMER, &tmp_u8);
		if (tmp_u8 != lat_timer) {
                        printk ("%d %d\n", tmp_u8, lat_timer);
			module_printk(KERN_ERR, "Latency timer of %d is not "
				      "supported (%x:%x)\n", lat_timer,
				      pdev->vendor, pdev->device);
			return -EINVAL;
		}
	}

        /* Configuring PCIe bridge (PEX), if any. */
	if (!bridge || (0x10b5 != bridge->vendor) ||
	    (0x8111 != bridge->device && 0x8112 != bridge->device))
                return 0;

	module_printk (KERN_INFO, "Setting PCIe bridge Cache Line Size to %d\n",
		       cacheline_bridge);
	if (set_cacheline (bridge, cacheline_bridge))
            return -EINVAL;

        return 0;
}

/*
Allocates memory for pmc_e1_controller structure, returns a pointer
to allocated strunture, NULL if allocation insucessful.	*/
static struct pmc_e1_controller * pmc_e1_alloc(void)
{
	struct pmc_e1_controller * tmp_dev;
	//create structure for pmc_device
	tmp_dev = kmalloc(sizeof(* tmp_dev), GFP_KERNEL);
	if(tmp_dev)
	{
		memset(tmp_dev,0, sizeof(*tmp_dev));
	}
	return tmp_dev;
}

static int ptmc_e1_init_one (struct pci_dev *pdev, const struct pci_device_id *ent)
{
	struct pmc_e1_controller * tmp_dev;
	int err=0;
        u8 lat_timer;

	if (pci_enable_device(pdev))
	    return -EIO;
	if (device_number > (PMC_E18_NR-1))
	    return -EBUSY;

	err = tune_pci (pdev);
	if (err)
	    return err;

	tmp_dev = pmc_e1_alloc();
	if (!tmp_dev)
	{
		module_printk(KERN_ERR, "Failed to allocate device.\n");
		return -ENOMEM;
	}
	/* TODO: move to prepare_io */
       	tmp_dev->dev_number = device_number;
        sprintf(tmp_dev->name,"quasar%d", device_number);
	pci_read_config_byte(pdev, PCI_REVISION_ID, &lat_timer);
	switch(lat_timer)
	{
	case 0xa2:
		tmp_dev->e1_amount = 1;
		break;
	case 0xa4:
		tmp_dev->e1_amount = 3;
		break;
	case 0xa8:
		tmp_dev->e1_amount = 7; //[0:7]
		break;
	default:
		tmp_dev->e1_amount = 7; //assume old revision device 8 ports only.
		break;
	}

	//bind with pci device;
	tmp_dev->pci_device = pdev;
	//add to list of controllers, I don't remember why should I use this list, just part of old code from old project.
	list_add_tail(&tmp_dev->dev_node, &dev_list_head);
      // printk("pciaddr = 0x%lx, irq = %d, ptmc_dev = %p \n", pci_resource_start(pdev,0), pdev->irq, tmp_dev);
	if (request_mem_region (pci_resource_start(pdev,0), WIN_SIZE, tmp_dev->name) == NULL)
	{
	    module_printk (KERN_ERR, "I/O resource 0x%x @ 0x%llx busy\n",
		    WIN_SIZE, (u64) pci_resource_start(pdev,0));
	    goto fail_req_bar0;
	}
	if (request_mem_region (pci_resource_start (pdev, 1), WIN_SIZE, tmp_dev->name) == NULL)
	{
	    module_printk (KERN_ERR, "I/O resource 0x%x @ 0x%llx busy\n",
		    WIN_SIZE, (u64) pci_resource_start(pdev,1));
	    goto fail_req_bar1;
	}
	tmp_dev->bar0 = ioremap_nocache(pci_resource_start(pdev,0), WIN_SIZE);
	if (!tmp_dev->bar0)	{
	    module_printk(KERN_ERR, "Can't map 0x%x @ 0x%llx\n",
		   WIN_SIZE, (u64) pci_resource_start(pdev,0));
	    goto fail_map_bar0;
	}
	tmp_dev->bar1 = ioremap_nocache(pci_resource_start(pdev,1), WIN_SIZE);
	if (!tmp_dev->bar1)	{
	    module_printk(KERN_ERR, "Can't map 0x%x @ 0x%llx\n",
		   WIN_SIZE, (u64) pci_resource_start(pdev,1));
	    goto fail_map_bar1;
	}
	ptmc_sysfs_create (tmp_dev);
        
	module_printk(KERN_INFO, "Registering interrupt %d with flags 0x%x\n",
		     pdev->irq, irqflags);
	if (request_irq(pdev->irq, ptmc_e1_interrupt_heandler,
			irqflags, tmp_dev->name, (void *) tmp_dev) < 0)
	{
	    module_printk(KERN_ERR, "Can't register interrupt: %d\n", pdev->irq);
	    goto fail_reg_irq;
	}
        cdev_init(&tmp_dev->dev_file, &pmc_e1_fops);
	tmp_dev->dev_file.owner = THIS_MODULE;
	tmp_dev->dev_file.ops = &pmc_e1_fops;
	if (cdev_add(&tmp_dev->dev_file, MKDEV(MAJOR(dev_file_glob),  device_number), 1) )
	{
	    module_printk(KERN_ERR, "Can't register device file for: %d\n", device_number);
	    goto fail_req_file;
	}
	device_number++; //for non-hot plug systems, for hot-plug system we probably should change system of numbering...
	spin_lock_init(&tmp_dev->dev_lock);
	//mark module as initialised
	tmp_dev->registered = 1;

	//Reset and initialise board;

	//initialising board registers
	init_ptmc_regs(tmp_dev);

	//init framer registers
	init_framer_regs(tmp_dev);

	//initialising time-slots maping table.
	init_switch_map(tmp_dev);

	//DAHDI
	module_printk(KERN_INFO, "Quasar board with %d ports found, dev id: %d, hw_fw_rev: 0x%x, (pci slot %d)\n",
		      tmp_dev->e1_amount + 1, tmp_dev->dev_number, tmp_dev->bar0[0x0D], PCI_SLOT(pdev->devfn)+1);
	err = e12zap_register(tmp_dev);
	if(err)
	    module_printk(KERN_ERR, "Quasar dahdi registration error.\n");
        return 0;
	//Clean up all resources, but leave in list.
fail_req_file: free_irq(pdev->irq, (void *) tmp_dev);

fail_reg_irq:
        ptmc_sysfs_delete (tmp_dev);
	iounmap((void *)tmp_dev->bar1);
fail_map_bar1: iounmap((void *)tmp_dev->bar0);
fail_map_bar0: release_mem_region(pci_resource_start(pdev, 1), WIN_SIZE);//unreg 1 bar;
fail_req_bar1: release_mem_region(pci_resource_start(pdev, 0), WIN_SIZE);//unreg 0 bar;
fail_req_bar0: tmp_dev->registered = 0;
return -EBUSY;
}

static void ptmc_e1_remove_one (struct pci_dev *pdev)
{
    struct list_head *pos, *n;
	struct pmc_e1_controller *tmp_dev=NULL;
	list_for_each_safe(pos, n, &dev_list_head)
	{
		if (list_entry(pos,struct pmc_e1_controller,dev_node)->pci_device == pdev)
		{
			tmp_dev = list_entry(pos, struct pmc_e1_controller,dev_node);
		}
	}
	if(tmp_dev)
	{
		if(tmp_dev->registered)
		{
			e12zap_unregister(tmp_dev); //DAHDI
			free_irq(pdev->irq, (void *) tmp_dev);
			cdev_del(&tmp_dev->dev_file);
			ptmc_sysfs_delete (tmp_dev);

			iounmap((void *)tmp_dev->bar0);
			release_mem_region (pci_resource_start(pdev, 0), WIN_SIZE);
			iounmap((void *)tmp_dev->bar1);
			release_mem_region (pci_resource_start(pdev, 1), WIN_SIZE);

		}
		//unlink from device's list
		list_del(&tmp_dev->dev_node);
		kfree(tmp_dev);
	}
}

static struct pci_driver ptmc_e1_driver = {
        name:		DRV_NAME,
	probe:		ptmc_e1_init_one,
	remove:		ptmc_e1_remove_one,
	id_table:	ptmc_e1_tbl,
};

static int __init ptmc_e1_init_module(void)
{
	int result, error;
	module_printk(KERN_INFO, "Quasar E1 interface board driver, Parabel Ltd. http://parabel.ru, "
		      "Version: %s\n", QUASAR_VERSION);

	if (PMC_E1_MAJOR)
	{
		dev_file_glob = MKDEV(PMC_E1_MAJOR,0);
		result = register_chrdev_region(dev_file_glob, PMC_E18_NR, DRV_NAME);
	}
	else
	{
		result = alloc_chrdev_region(&dev_file_glob, 0, PMC_E18_NR, DRV_NAME);
	}
	if (result < 0)
	{
		module_printk(KERN_ERR, "Can't register major number for Quasar driver.\n");
		return -EBUSY;
	}
#ifdef PMC_E1_UDEV
       pmc_e1_class = class_create(THIS_MODULE, DRV_NAME);
	if (IS_ERR(pmc_e1_class))
	{
		module_printk(KERN_ERR, "Can't create class " DRV_NAME ".\n");
		error=-PTR_ERR(pmc_e1_class);
		goto fail_reg_class;
	}
#endif
	result = pci_register_driver(&ptmc_e1_driver);
	if(result >= 0)
	{
		return 0;
	}
	else
	{
		module_printk(KERN_ERR, "PCI registration for Quasar board failed!");
		error=-EBUSY;
	}
	pci_unregister_driver(&ptmc_e1_driver);

#ifdef PMC_E1_UDEV
	class_destroy(pmc_e1_class);
fail_reg_class:
#endif


	unregister_chrdev_region(dev_file_glob, PMC_E18_NR);
	return error;
}

static void __exit ptmc_e1_cleanup_module(void)
{
	pci_unregister_driver (&ptmc_e1_driver);
	unregister_chrdev_region(dev_file_glob, PMC_E18_NR);
	module_printk(KERN_INFO, "Quasar E1 interface board driver unloaded.\n");

#ifdef PMC_E1_UDEV
	class_destroy(pmc_e1_class);
#endif

    return;
}

module_init(ptmc_e1_init_module);
module_exit(ptmc_e1_cleanup_module);

module_param(cacheline, int, 0600);
MODULE_PARM_DESC(cacheline, "PCI Cache Line Size (in units of DWORDs).");
module_param(cacheline_bridge, int, 0600);
MODULE_PARM_DESC(cacheline_bridge, "PCIe bridge Cache Line Size (DWORDs).");
module_param(latency, int, 0400);
MODULE_PARM_DESC(latency, "PCI BUS latency timer.");
module_param(irqflags, int, 0600);
MODULE_PARM_DESC(irqflags, "Interrupt handler flags (passed to request_irq).");

MODULE_AUTHOR("Khelik Mikhail <mixel.net@gmail.com>, Pavel Selivanov <pavel[AT]parabel.ru>");
MODULE_DESCRIPTION("Quasar E1 interface board driver");
MODULE_LICENSE("GPL");
MODULE_VERSION (QUASAR_VERSION);
